/*
  Frequency Counter

  T. Nakagawa

  2010-03-16
*/


#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include "uart.h"
#include "14seglcd.h"


#define MODE_L 0	/* low frequency mode (-5Mhz, gate time=1s, resolution=1Hz) */
#define MODE_M 1	/* middle frequency mode (-300Hz, gate time=0.64s, resolution=100Hz) */
#define MODE_H 2	/* high frequency mode (-1300Hz, gate time=0.256s, resolution=1000Hz) */


volatile static uint8_t Mode;
volatile static uint16_t Cnt;


/*
  [hύX
*/
static void setmode(void) {
  if (Mode == MODE_L) {
    PORTD &= ~_BV(PD3);
  } else if (Mode == MODE_M) {
    PORTD |= _BV(PD3);
    PORTD |= _BV(PD4);	/* MB506:1/64 */
  } else {
    PORTD |= _BV(PD3);
    PORTD &= ~_BV(PD4);	/* MB506:1/256 */
  }

  return;
}


/*
  L[͂̏([hؑ)
*/
static int8_t keysense(void) {
  static uint8_t prv = 0xff;
  static uint8_t cnt = 0;
  uint8_t tmp;

  /* 100HzɃL[͂TvO */
  if (cnt-- != 0) return 0;
  cnt = 62;

  tmp = prv;
  prv = PIND;
  if (((tmp ^ prv) & ~prv & _BV(PD2)) != 0) {	/* XCb`ꂽ */
    Mode++;
    if (Mode > 2) Mode = 0;
    setmode();

    return 1;
  }

  return 0;
}


/*
  [hɉĎgJEg
*/
static uint32_t cntfreq(void) {
  uint8_t ovf;
  uint16_t loop;
  uint16_t off, cnt;
  uint32_t frq;

restart:

  /*  */
  ovf = 0;
  if (Mode == MODE_L) {
    loop = 6250;	/* (1.0 / (1.0 / (12.8e6 / 8 / 256))) Q[g^C=1s */
  } else if (Mode == MODE_M) {
    loop = 4000;	/* (0.64 / (1.0 / (12.8e6 / 8 / 256))) Q[g^C=0.64s */
  } else {
    loop = 1600;	/* (0.256 / (1.0 / (12.8e6 / 8 / 256))) Q[g^C=0.256s */
  }
  TCNT1 = 0;
  TIFR |= _BV(TOV1);

  /* JñJEgl𓾂(I[o[t[͋NȂƂ݂Ȃ) */
  asm volatile(
    "sei\n\t"
    "sleep\n\t"
    "cli\n\t"
  ::);
  off = Cnt;

  /* Q[gԃJEg */
  do {
    /* 荞݂҂ */
    asm volatile(
      "sei\n\t"
      "sleep\n\t"
      "cli\n\t"
    ::);
    /* ȍ~̏8*256NbN(̊荞݂܂)ZԂɍsȂƂȂ */
    /* I[o[t[̏ */
    if (bit_is_set(TIFR, TOV1) && Cnt < 0x8000) {	/* ۑCntOVFNꍇ */
      ovf++;	/* I[o[t[12.8MHz/2/65536N肤 */
      TIFR |= _BV(TOV1);
    }
    /* L[ */
    if (keysense()) goto restart;
  } while (--loop != 0);
  cnt = Cnt;

  frq = 65536 * ovf + cnt - off;

  return frq;
}


/*
  [hɉĎgLCDɕ\ & UARTɏo
*/
static void display(uint32_t frq) {
  int8_t i, j;
  uint8_t zero;
  uint8_t pos;

  lcd_buf[8] &= ~0xf0;
  if (Mode == MODE_L) {
    zero = 0;
    pos = 4;
    lcd_buf[8] |= 0x40;
  } else if (Mode == MODE_M) {
    zero = 2;
    pos = 6;
    lcd_buf[8] |= 0x20;
  } else {
    zero = 3;
    pos = 7;
    lcd_buf[8] |= 0x10;
  }

  for (i = 7; i >= 0; i--) {
    if (i == pos) {
      lcd_buf[i] = LCD_K;
    } else {
      lcd_buf[i] = (uint8_t)(frq % 10);
      frq /= 10;
      if (frq == 0) {
        /* UARTɏo */
        for (j = i; j <= 7; j++) {
          uart_putchar('0' + lcd_buf[j]);
        }
        for (j = 0; j < zero; j++) uart_putchar('0');
        uart_putchar('\r');
        uart_putchar('\n');
        /* zero-suppress */
        if (i < pos) break;
      }
    }
  }

  for (i--; i >= 0; i--) {
    lcd_buf[i] = LCD_SPACE;
  }

  lcd_buf[8] ^= 0x04;	/* Gate time indicator */

  lcd_refresh();

  return;
}


int main(void) {
  uint32_t f;

  /* |[g */
  PORTB |= (_BV(PB5) | _BV(PB6) | _BV(PB7));
  DDRD |= (_BV(PD1) | _BV(PD3) | _BV(PD4));
  PORTD |= _BV(PD2);

  /* LCD */
  lcd_init();

  /* UART */
  uart_init();

  /* AiORp[^~ */
  ACSR |= _BV(ACD);

  /* ^C}EJE^ */
  TCCR0 = 0x02;	/* 1/(12.8MHz/8/256)secɊ荞݂𔭐 */
  TCCR1B |= 0x07;	/* O̗͂オJEg */
  TIMSK |= _BV(TOIE0);
  MCUCR |= _BV(SE);

  /*  */
  Mode = MODE_L;
  setmode();

  for (; ; ) {
    /* gJEg */
    f = cntfreq();

    /* LCD\ & UARTo */
    display(f);
  }

  return 0;
}


/*
  ^C}0I[o[t[荞
*/
ISR (TIMER0_OVF0_vect) {
  /* JE^ۑ */
  Cnt = TCNT1;
}
